/*
stk500boot.c  20030810

Copyright (c) 2003, Jason P. Kyle
All rights reserved.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Licence can be viewed at http://www.fsf.org/licenses/gpl.txt


Target = Atmel AVR m128,m64,m32,m16,m8,m162,m163,m169,m8515,m8535
ATmega161 has a very small boot block so isn't supported.

Tested with m128,m8,m163 - feel free to let me know how/if it works for you.
*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/boot.h>
//#include <avr/wdt.h>

#define	cbi(P,B)	(P) &= (~(1<<(B)))
#define	sbi(P,B)	(P) |= (1<<(B))
#define	inb(P)		(P)
#define	outb(P,V)		(P) = (V)

//#define F_CPU			3686400
//#define F_CPU			14745600
#define BAUD_RATE		115200

#define DECRYPT 0
#define ENCRYPT 1
//#define DES_ENCRYPTION

#define HW_VER		0x02
#define SW_MAJOR	0x02
#define SW_MINOR	0x0a


/* Adjust to suit whatever pin your hardware uses to enter the bootloader */
/*
#define BL_DDR DDRB
#define BL_PORT PORTB
#define BL_PIN PINB
#define BL PINB5
*/

/*
 * Charon II. Development Board bootloader entry point - CEBA 14.09.03
 */
#define BL_DDR DDRE
#define BL_PORT PORTE
#define BL_PIN PINE
#define BL PORTE6

//#define __AVR_ATmega128__

#define SIG1	0x1E	// Yep, Atmel is the only manufacturer of AVR micros.  Single source :(
#if defined __AVR_ATmega128__
	#define SIG2	0x97
	#define SIG3	0x02
	#define PAGE_SIZE	0x80U	//128 words
	#define UART0
//	#define UART1
#elif defined __AVR_ATmega64__
	#define SIG2	0x96
	#define SIG3	0x02
	#define PAGE_SIZE	0x80U	//128 words
	#define UART0
//	#define UART1
#elif defined __AVR_ATmega32__
	#define SIG2	0x95
	#define SIG3	0x02
	#define PAGE_SIZE	0x40U	//64 words
#elif defined __AVR_ATmega16__
	#define SIG2	0x94
	#define SIG3	0x03
	#define PAGE_SIZE	0x40U	//64 words
#elif defined __AVR_ATmega8__
	#define SIG2	0x93
	#define SIG3	0x07
	#define PAGE_SIZE	0x20U	//32 words
#elif defined __AVR_ATmega162__
	#define SIG2    0x94
	#define SIG3    0x04
	#define PAGE_SIZE	0x40U	//64 words
	#define UART0
//	#define UART1
#elif defined __AVR_ATmega163__
	#define SIG2	0x94
	#define SIG3	0x02
	#define PAGE_SIZE	0x40U	//64 words
#elif defined __AVR_ATmega169__
	#define SIG2	0x94
	#define SIG3	0x05
	#define PAGE_SIZE	0x40U	//64 words
#elif defined __AVR_ATmega8515__
	#define SIG2	0x93
	#define SIG3	0x06
	#define PAGE_SIZE	0x20U	//32 words
#elif defined __AVR_ATmega8535__
	#define SIG2	0x93
	#define SIG3	0x08
	#define PAGE_SIZE	0x20U	//32 words
#endif


void putch(char ch);
char getch(void);
void getNch(uint8_t);
void byte_response(uint8_t);
void nothing_response(void);


union address_union {
	uint16_t word;
	uint8_t  byte[2];
} address;

union length_union {
	uint16_t word;
	uint8_t  byte[2];
} length;

struct flags_struct {
	unsigned eeprom : 1;
	unsigned rampz  : 1;
} flags;

uint8_t buff[256];
uint8_t address_high;

uint8_t pagesz=0x80;

void (*app_start)(void) = 0x0000;

int main(void)
{
uint8_t ch,ch2;
uint16_t w;
#ifdef	BL_DDR
	cbi(BL_DDR,BL);
	sbi(BL_PORT,BL);
	asm volatile("nop\n\t");
	if(__LPM(0x0000) != 0xFF)	/* Don't start application if it isn't programmed yet */
	{		 
		/* Run application if key not pressed and no power on or external reset occurs */
		/* No reset happens means a jump from application program to bootloader occurs */
		if(bit_is_set(BL_PIN,BL) && ((MCUCSR & 1<<EXTRF) || (MCUCSR & 1<<PORF)))
		{
			MCUCSR &= ~(1<<EXTRF | 1<<PORF);	/* clear the reset flag */
			app_start();
		}
	}
#endif

#ifdef __AVR_ATmega163__
	UBRR = (uint8_t)(F_CPU/(BAUD_RATE*16L)-1);
	UBRRHI = (F_CPU/(BAUD_RATE*16L)-1) >> 8;
	UCSRA = 0x00;
	UCSRB = _BV(TXEN)|_BV(RXEN);
#elif defined UART0
	UBRR0L = (uint8_t)(F_CPU/(BAUD_RATE*16L)-1);
	UBRR0H = (F_CPU/(BAUD_RATE*16L)-1) >> 8;
	UCSR0A = 0x00;
	UCSR0C = 0x06;
	UCSR0B = _BV(TXEN0)|_BV(RXEN0);
#elif defined UART1
	UBRR1L = (uint8_t)(F_CPU/(BAUD_RATE*16L)-1);
	UBRR1H = (F_CPU/(BAUD_RATE*16L)-1) >> 8;
	UCSR1A = 0x00;
	UCSR1C = 0x06;
	UCSR1B = _BV(TXEN1)|_BV(RXEN1);
#else		// m8m,16,32,169,8515,8535
	UBRRL = (uint8_t)(F_CPU/(BAUD_RATE*16L)-1);
	UBRRH = (F_CPU/(BAUD_RATE*16L)-1) >> 8;
	UCSRA = 0x00;
	UCSRC = 0x06;
	UCSRB = _BV(TXEN)|_BV(RXEN);
#endif
//	putch('\0');
	if(boot_lock_fuse_bits_get(GET_LOCK_BITS ) & ~0xd7)
	{
		__boot_lock_bits_set(~0xd7);
	}
 for (;;) {
   ch = getch();
	if(ch=='0') {		// Hello is anyone home?
		nothing_response();
	}
		// Request programmer ID
	else if(ch=='1') {		//Yes i've heard of the switch statement, a bunch of else if's -> smaller code
		if (getch() == ' ') {
			putch(0x14);
			putch('A');	//Not using PROGMEM string due to boot block in m128 being beyond 64kB boundry
			putch('V');	//Would need to selectively manipulate RAMPZ, and it's only 9 characters anyway so who cares.
			putch('R');
			putch(' ');
			putch('I');
			putch('S');
			putch('P');
			putch(0x10);
		}
	}
	else if(ch=='3'){
		getch();
		nothing_response();
	}
	else if(ch=='@') {		// AVR ISP/STK500 board commands  DON'T CARE so default nothing_response
		ch2 = getch();
		if (ch2>0x85) getch();
		nothing_response();
	}
	else if(ch=='A') {		// AVR ISP/STK500 board requests
		ch2 = getch();
		if(ch2==0x80) byte_response(HW_VER);		// Hardware version
		else if(ch2==0x81) byte_response(SW_MAJOR);	// Software major version
		else if(ch2==0x82) byte_response(SW_MINOR);	// Software minor version
		else if(ch2==0x98) byte_response(0x03);		// Unknown but seems to be required by avr studio 3.56
		else byte_response(0x00);					// Covers various unnecessary responses we don't care about
	}
	else if(ch=='B') {		// Device Parameters  DON'T CARE, DEVICE IS FIXED
		getNch(20);
		nothing_response();
	}
	else if(ch=='E') {		// Parallel programming stuff  DON'T CARE
		getNch(4);
		nothing_response();
	}
	else if(ch=='P') {		// Enter programming mode
		nothing_response();
	}
	else if(ch=='Q') {		// Leave programming mode
		nothing_response();
		//app_start();
	}
	else if(ch=='R') {		// Erase device, don't care as we will erase one page at a time anyway.
		nothing_response();
	}
	else if(ch=='U') {		//Set address, little endian. EEPROM in bytes, FLASH in words
							//Perhaps extra address bytes may be added in future to support > 128kB FLASH.
							//This might explain why little endian was used here, big endian used everywhere else.
		address.byte[0] = getch();
		address.byte[1] = getch();
		nothing_response();
	}
	else if(ch=='V') {		// Universal SPI programming command, disabled.  Would be used for fuses and lock bits.
		ch = getch();
		ch2 = getch();
		if(ch == 0x50 && ch2 ==0x00)
		{
			ch = boot_lock_fuse_bits_get(GET_LOW_FUSE_BITS );
		}
		else if(ch == 0x58 && ch2 == 0x08)
		{
			ch = boot_lock_fuse_bits_get(GET_HIGH_FUSE_BITS );
		}
		else if(ch == 0x50 && ch2 == 0x08)
		{
			ch = boot_lock_fuse_bits_get(GET_EXTENDED_FUSE_BITS );
		}
		else if(ch == 0x58 && ch2 == 0x00)
		{
			ch = boot_lock_fuse_bits_get(GET_LOCK_BITS );
		}
		else
		{
			ch = 0x00;
		}
		getNch(2);
		byte_response(ch);
	}
	else if(ch=='d') {		// Write memory, length is big endian and is in bytes
		length.byte[1] = getch();
		length.byte[0] = getch();
		flags.eeprom = 0;
		if (getch() == 'E') flags.eeprom = 1;
		for (w=0;w<length.word;w++) {
		  buff[w] = getch();	// Store data in buffer, can't keep up with serial data stream whilst programming pages
		}
		if (getch() == ' ') {
			if (flags.eeprom) {		//Write to EEPROM one byte at a time
				for(w=0;w<length.word;w++) {
					eeprom_write_byte( (char *)(&address.word),buff[w]);
					address.word++;
				}
			}
			else {					//Write to FLASH one page at a time
				if (address.byte[1]>127) address_high = 0x01;	//Only possible with m128, m256 will need 3rd address byte. FIXME
				else address_high = 0x00;
#ifdef __AVR_ATmega128__
				RAMPZ = address_high;
#endif
				address.word = address.word << 1;	//address * 2 -> byte location
//				if ((length.byte[0] & 0x01) == 0x01) length.word++;	//Even up an odd number of bytes
				if ((length.byte[0] & 0x01)) length.word++;	//Even up an odd number of bytes
				cli();									//Disable interrupts, just to be sure
				while(bit_is_set(EECR,EEWE));			//Wait for previous EEPROM writes to complete
				asm volatile("clr	r17				\n\t"	//page_word_count
							 "lds	r30,address		\n\t"	//Address of FLASH location (in bytes)
							 "lds	r31,address+1	\n\t"
							 "ldi	r28,lo8(buff)		\n\t"	//Start of buffer array in RAM
							 "ldi	r29,hi8(buff)		\n\t"
							 "lds	r24,length		\n\t"	//Length of data to be written (in bytes)
							 "lds	r25,length+1	\n\t"
							 "length_loop:			\n\t"	//Main loop, repeat for number of words in block
							 "cpi	r17,0x00	\n\t"	//If page_word_count=0 then erase page
							 "brne	no_page_erase	\n\t"
							 "wait_spm1:			\n\t"
							 "lds	r16,%0		\n\t"	//Wait for previous spm to complete
							 "andi	r16,1\n\t"
							 "cpi	r16,1\n\t"
							 "breq	wait_spm1\n\t"
							 "ldi	r16,0x03		\n\t"	//Erase page pointed to by Z
							 "sts	%0,r16		\n\t"
							 "spm					\n\t"
#ifdef __AVR_ATmega163__
	 						 ".word 0xFFFF					\n\t"
							 "nop							\n\t"
#endif
							 "wait_spm2:			\n\t"
							 "lds	r16,%0		\n\t"	//Wait for previous spm to complete
							 "andi	r16,1\n\t"
							 "cpi	r16,1\n\t"
							 "breq	wait_spm2\n\t"

							 "ldi	r16,0x11				\n\t"	//Re-enable RWW section
					 		 "sts	%0,r16				\n\t"
					 		 "spm							\n\t"
#ifdef __AVR_ATmega163__
	 						 ".word 0xFFFF					\n\t"
							 "nop							\n\t"
#endif
							 "no_page_erase:		\n\t"
							 "ld	r0,Y+			\n\t"		//Write 2 bytes into page buffer
							 "ld	r1,Y+			\n\t"

							 "wait_spm3:			\n\t"
							 "lds	r16,%0		\n\t"	//Wait for previous spm to complete
							 "andi	r16,1\n\t"
							 "cpi	r16,1\n\t"
							 "breq	wait_spm3\n\t"
							 "ldi	r16,0x01		\n\t"	//Load r0,r1 into FLASH page buffer
							 "sts	%0,r16		\n\t"
							 "spm					\n\t"

							 "inc	r17				\n\t"	//page_word_count++
							 "cpi r17,%1	\n\t"
							 "brlo	same_page		\n\t"	//Still same page in FLASH
							 "write_page:			\n\t"
							 "clr	r17				\n\t"	//New page, write current one first
							 "wait_spm4:			\n\t"
							 "lds	r16,%0		\n\t"	//Wait for previous spm to complete
							 "andi	r16,1\n\t"
							 "cpi	r16,1\n\t"
							 "breq	wait_spm4\n\t"
#ifdef __AVR_ATmega163__
							 "andi	r30,0x80		\n\t"	// m163 requires Z6:Z1 to be zero during page write
#endif
							 "ldi	r16,0x05		\n\t"	//Write page pointed to by Z
							 "sts	%0,r16		\n\t"
							 "spm					\n\t"
#ifdef __AVR_ATmega163__
	 						 ".word 0xFFFF					\n\t"
							 "nop							\n\t"
							 "ori	r30,0x7E		\n\t"		// recover Z6:Z1 state after page write (had to be zero during write)
#endif
							 "wait_spm5:			\n\t"
							 "lds	r16,%0		\n\t"	//Wait for previous spm to complete
							 "andi	r16,1\n\t"
							 "cpi	r16,1\n\t"
							 "breq	wait_spm5\n\t"
							 "ldi	r16,0x11				\n\t"	//Re-enable RWW section
					 		 "sts	%0,r16				\n\t"
					 		 "spm							\n\t"
#ifdef __AVR_ATmega163__
	 						 ".word 0xFFFF					\n\t"
							 "nop							\n\t"
#endif
							 "same_page:			\n\t"
							 "adiw	r30,2			\n\t"	//Next word in FLASH
							 "sbiw	r24,2			\n\t"	//length-2
							 "breq	final_write		\n\t"	//Finished
							 "rjmp	length_loop		\n\t"
							 "final_write:			\n\t"
							 "cpi	r17,0			\n\t"
							 "breq	block_done		\n\t"
							 "adiw	r24,2			\n\t"	//length+2, fool above check on length after short page write
							 "rjmp	write_page		\n\t"
							 "block_done:			\n\t"
							 "clr	__zero_reg__	\n\t"	//restore zero register
							 : "=m" (SPMCR) : "M" (PAGE_SIZE) : "r0","r16","r17","r24","r25","r28","r29","r30","r31");

/* Should really add a wait for RWW section to be enabled, don't actually need it since we never */
/* exit the bootloader without a power cycle anyhow */
			}
			putch(0x14);
			putch(0x10);
		}
	}
	else if(ch=='t') {		//Read memory block mode, length is big endian.
		length.byte[1] = getch();
		length.byte[0] = getch();
#if defined __AVR_ATmega128__
		if (address.word>0x7FFF) flags.rampz = 1;		// No go with m256, FIXME
		else flags.rampz = 0;
#endif
		if (getch() == 'E') flags.eeprom = 1;
		else {
			flags.eeprom = 0;
			address.word = address.word << 1;	//address * 2 -> byte location
		}
		if (getch() == ' ') {		// Command terminator
			putch(0x14);
			for (w=0;w < length.word;w++) {		// Can handle odd and even lengths okay
				if (flags.eeprom) {	// Byte access EEPROM read
					putch(eeprom_read_byte((char *)(&address.word)));
					address.word++;
				}
				else {
					if (!flags.rampz) putch(__LPM(address.word));
#if defined __AVR_ATmega128__
					else putch(__LPM(address.word + 0x10000));	// Hmmmm, yuck  FIXME when m256 arrvies
#endif
					address.word++;
				}
			}
			putch(0x10);
		}
	}
	else if(ch=='u') {		// Get device signature bytes
		if (getch() == ' ') {
			putch(0x14);
			putch(SIG1);
			putch(SIG2);
			putch(SIG3);
			putch(0x10);
		}
	}
	else if(ch=='v') {		// Read oscillator calibration byte
		byte_response(0x00);
	}
  }
}

void putch(char ch)
{
#ifdef UART0
	while (!(inb(UCSR0A) & _BV(UDRE0)));
	outb(UDR0,ch);
#elif defined UART1
	while (!(inb(UCSR1A) & _BV(UDRE1)));
	outb(UDR1,ch);
#else		// m8,16,32,169,8515,8535,163
	while (!(inb(UCSRA) & _BV(UDRE)));
	outb(UDR,ch);
#endif
}

char getch(void)
{
#ifdef UART0
	while(!(inb(UCSR0A) & _BV(RXC0)));
	return (inb(UDR0));
#elif defined UART1
	while(!(inb(UCSR1A) & _BV(RXC1)));
	return (inb(UDR1));
#else		// m8,16,32,169,8515,8535,163
	while(!(inb(UCSRA) & _BV(RXC)));
	return (inb(UDR));
#endif
}

void getNch(uint8_t count)
{
uint8_t i;
	for(i=0;i<count;i++) {
#ifdef UART0
		while(!(inb(UCSR0A) & _BV(RXC0)));
		inb(UDR0);
#elif defined UART1
		while(!(inb(UCSR1A) & _BV(RXC1)));
		inb(UDR1);
#else		// m8,16,32,169,8515,8535,163
		while(!(inb(UCSRA) & _BV(RXC)));
		inb(UDR);
#endif
	}
}

void byte_response(uint8_t val)
{
	if (getch() == ' ') {
		putch(0x14);
		putch(val);
		putch(0x10);
	}
}

void nothing_response(void)
{
	if (getch() == ' ') {
		putch(0x14);
		putch(0x10);
	}
}

